package org.expeditee.actions;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.expeditee.agents.Agent;
import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameUtils;
import org.expeditee.io.Conversion;
import org.expeditee.io.Logger;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;

/**
 * The Action class is used to launch KMS Actions as well as launching JAGs.
 * 
 * This class checks all class files in the same directory, and reads in and
 * adds all the methods from them. The methods are stored in a Hashtable so that
 * the lowercase method names can be mapped to the correctly capatilized method
 * names (to provide case-insensitivity)
 */
public class Actions {

	// the currently running agent (if there is one)
	private static Agent _Agent = null;

	// maps lower case method names to the method
	private static HashMap<String, Method> _Actions = new HashMap<String, Method>();

	// maps lower case JAG class names to capitalized JAG names
	private static HashMap<String, String> _JAGs = new HashMap<String, String>();

	public static final String ROOT_PACKAGE = "org.expeditee.";
	
	// Package and class file locations
	private static final String ACTIONS_PACKAGE = ROOT_PACKAGE + "actions.";

	private static final String AGENTS_PACKAGE = ROOT_PACKAGE + "agents.";
	
	private static final String NAVIGATIONS_CLASS = ROOT_PACKAGE + "actions.NavigationActions";
	
	public static Class[] getClasses(String pckgname)
			throws ClassNotFoundException {
		ArrayList<Class> classes = new ArrayList<Class>();
		// Get a File object for the package
		File directory = null;
		String path = pckgname.replace('.', '/');
		try {
			ClassLoader cld = Thread.currentThread().getContextClassLoader();
			if (cld == null) {
				throw new ClassNotFoundException("Can't get class loader.");
			}
			URL resource = cld.getResource(path);
			if (resource == null) {
				throw new ClassNotFoundException("No resource for " + path);
			}
			directory = new File(resource.getFile());
		} catch (NullPointerException x) {
			throw new ClassNotFoundException(pckgname + " (" + directory
					+ ") does not appear to be a valid package");
		}

		int splitPoint = directory.getPath().indexOf('!');
		if (splitPoint > 0) {
			try {
				String jarName = directory.getPath().substring(
						"file:".length(), splitPoint);
				// Windows HACK
				if (jarName.indexOf(":") >= 0)
					jarName = jarName.substring(1);

				if (jarName.indexOf("%20") > 0) {
					jarName = jarName.replace("%20", " ");
				}

				JarFile jarFile = new JarFile(jarName);

				Enumeration entries = jarFile.entries();
				while (entries.hasMoreElements()) {
					ZipEntry entry = (ZipEntry) entries.nextElement();
					String className = entry.getName();
					if (className.startsWith(path)) {
						if (className.endsWith(".class")
								&& !className.contains("$")) {
							classes.add(Class.forName(className.substring(0,
									className.length() - 6).replace('/', '.')));
						}
					}
				}
				jarFile.close();

			} catch (Exception e) {
				e.printStackTrace();
			}

		} else {

			if (directory.exists()) {
				// Get the list of the files contained in the package
				String[] files = directory.list();
				for (int i = 0; i < files.length; i++) {
					// we are only interested in .class files
					if (files[i].endsWith(".class") && !files[i].contains("$")
							&& !files[i].equals("Actions.class")) {
						// removes the .class extension
						classes
								.add(Class.forName(pckgname
										+ files[i].substring(0, files[i]
												.length() - 6)));
					}
				}
			} else {
				throw new ClassNotFoundException(pckgname + " (" + directory
						+ ") does not appear to be a valid package");
			}
		}
		Class[] classesA = new Class[classes.size()];
		classes.toArray(classesA);
		return classesA;
	}

	/**
	 * Clears out the Action and JAG Hashtables and refills them. Normally this
	 * is only called once when the system starts.
	 */
	public static void Init() {
		Class[] classes;
		try {
			classes = getClasses(AGENTS_PACKAGE);

			for (int i = 0; i < classes.length; i++) {
				String name = classes[i].getSimpleName();
				// maps lower case name to correct capitalised name
				_JAGs.put(name.toLowerCase(), name);
			}
		} catch (Exception e) {
			System.out
					.println("You must have Java 1.5 or higher to run Expeditee");
			System.out.println(e.getMessage());
		}
		try {
			classes = getClasses(ACTIONS_PACKAGE);

			for (int i = 0; i < classes.length; i++) {
				String name = classes[i].getSimpleName();
				// read in all the methods from the class
				try {
					// System.out.println(name);
					LoadMethods(Class.forName(ACTIONS_PACKAGE + name));
				} catch (ClassNotFoundException e) {
					Logger.Log(e);
					e.printStackTrace();
				}
			}
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}

	/**
	 * Loads all the Methods that meet the requirements checked by MethodCheck
	 * into the hashtable.
	 * 
	 * @param c
	 *            The Class to load the Methods from.
	 */
	private static void LoadMethods(Class c) {
		// list of methods to test
		Method[] toLoad = c.getMethods();

		for (Method m : toLoad) {
			// only allow methods with the right modifiers
			if (MethodCheck(m)) {
				if (!(_Actions.containsKey(m.getName().toLowerCase())))
					_Actions.put(m.getName().toLowerCase(), m);
				else {
					int i = 0;
					while (_Actions.containsKey(m.getName().toLowerCase() + i))
						i++;

					_Actions.put(m.getName().toLowerCase() + i, m);
				}

			}
		}
	}

	/**
	 * Checks if the given Method corresponds to the restrictions of Action
	 * commands, namely: Declared (not inherited), Public, and Static, with a
	 * void return type.
	 * 
	 * @param m
	 *            The Method to check
	 * @return True if the Method meets the above conditions, false otherwise.
	 */
	private static boolean MethodCheck(Method m) {
		// the return type must be void
		if (m.getReturnType() != void.class)
			return false;

		int mods = m.getModifiers();

		// check the method is declared (not inherited)
		if ((mods & Method.DECLARED) != Method.DECLARED)
			return false;

		// check the method is public
		if ((mods & Modifier.PUBLIC) != Modifier.PUBLIC)
			return false;

		// check the method is static
		if ((mods & Modifier.STATIC) != Modifier.STATIC)
			return false;

		// if we have not returned yet, then the tests have all passed
		return true;
	}

	/**
	 * Performs the given KMS action command. The source Frame and Item are
	 * given because they are required by some actions. Note that the source
	 * frame does not have to be the Item's parent Frame.
	 * 
	 * @param source
	 *            The Frame that the action should apply to
	 * @param launcher
	 *            The Item that has the action assigned to it
	 * @param command
	 *            The action to perform
	 */
	public static void PerformAction(Frame source, Item launcher, String command) {
		FrameIO.SaveFrame(source, false);
		
		// separate method name and parameter names
		String mname = getName(command);
		if (command.length() > mname.length())
			command = command.substring(mname.length() + 1);
		else
			command = "";

		// Strip off the @ from annotation items
		if (mname.startsWith("@"))
			mname = mname.substring(1);

		mname = mname.trim().toLowerCase();

		// check for protection on frame
		if (ItemUtils.ContainsTag(source.getItems(), "@No" + mname)) {
			FrameGraphics.DisplayMessage("Frame is protected by @No" + mname
					+ " tag.");
			return;
		}

		// retrieve methods that match the name
		Method toRun = _Actions.get(mname);

		// if this is not the name of a method, it may be the name of an agent
		if (toRun == null) {
			LaunchAgent(mname, command, source);
			return;
		}

		// Need to save the frame if we are navigating away from it so we dont
		// loose changes
		if (toRun.getDeclaringClass().getName().equals(
				 NAVIGATIONS_CLASS)) {
			FrameIO.SaveFrame(DisplayIO.getCurrentFrame());
		}

		// if there are duplicate methods with the same name
		List<Method> possibles = new LinkedList<Method>();
		possibles.add(toRun);
		int i = 0;
		while (_Actions.containsKey(mname + i)) {
			possibles.add(_Actions.get(mname + i));
			i++;
		}

		for (Method possible : possibles) {

			// try first with the launching item as a parameter

			// run method
			try {
				// convert parameters to objects and get the method to invoke
				Object[] parameters = CreateObjects(possible, source, launcher,
						command);

				possible.invoke(null, parameters);
				return;
			} catch (Exception e) {
				Logger.Log(e);
				e.printStackTrace();
			}
		}
	}

	/**
	 * Launches an agent with the given name, and passes in the given parameters
	 * 
	 * @param name
	 *            The name of the JAG to load
	 * @param parameters
	 *            The parameters to pass to the JAG
	 * @param source
	 *            The starting Frame that the JAG is being launched on
	 */
	private static void LaunchAgent(String name, String parameters, Frame source) {
		// save the current frame (if necesssary)
		FrameUtils.LeavingFrame(source);

		try {
			// check for stored capitalisation
			if (_JAGs.containsKey(name.toLowerCase())) {
				name = _JAGs.get(name.toLowerCase());
			} else if (name.toLowerCase().endsWith("tree")) {
				parameters = name.substring(0, name.length() - "tree".length())
						+ " " + parameters;
				name = "WriteTree";
			} else if (name.toLowerCase().endsWith("frame")) {
				parameters = name
						.substring(0, name.length() - "frame".length())
						+ " " + parameters;
				name = "WriteFrame";
			}

			// load the JAG class
			Class agentClass = Class.forName(AGENTS_PACKAGE + name);

			// get the constructor for the JAG class
			Constructor con = null;
			Constructor constructors[] = agentClass.getConstructors();
			Object[] params = null;

			// determine correct parameters for constructor
			for (Constructor c : constructors) {
				if (parameters.length() > 0
						&& c.getParameterTypes().length == 1) {
					con = c;
					params = new String[1];
					params[0] = parameters;
					break;
				} else if (c.getParameterTypes().length == 0 && con == null)
					con = c;
			}

			// if there is no constructor, return
			if (con == null) {
				FrameGraphics.DisplayMessage("Invalid parametres for agent.");
				//System.out.println("Constructor not found...");
				return;
			}

			// create the JAG
			_Agent = (Agent) con.newInstance(params);

			Thread t = new Thread(_Agent);
			t.setPriority(Thread.MIN_PRIORITY);

			// check for errors during initialisation
			if (!_Agent.initialise(source)) {
				FrameGraphics.ErrorMessage("Error initialising agent: " + name);
				return;
			}

			_Agent.setStartFrame(source);

			t.start();

			// if there is a completion frame, then display it to the user
			if (_Agent.hasResultFrame()) {
				Frame next = _Agent.getResultFrame();
				FrameUtils.DisplayFrame(next, true);
			}

		} catch (ClassNotFoundException cnf) {
			FrameGraphics.ErrorMessage("Error: '" + name
					+ "' is not an action statement or Agent.");
		} catch (Exception e) {
			FrameGraphics.ErrorMessage("Error creating Agent: '" + name + "'");
			System.out.println("Agent set to Null.");
			_Agent = null;
			e.printStackTrace();
			Logger.Log(e);
		}

		return;
	}

	/**
	 * Used to determine if the previously launched agent is still executing.
	 * 
	 * @return True if the last Agent is still executing, False otherwise.
	 */
	public static boolean isAgentRunning() {
		if (_Agent != null)
			return _Agent.isRunning();

		return false;
	}

	/**
	 * Stops the currently running Agent (If there is one) by calling
	 * Agent.stop(). Note: This may not stop the Agent immediately, but the
	 * Agent should terminate as soon as it is safe to do so.
	 */
	public static void stopAgent() {
		if (_Agent != null) {
			FrameGraphics.DisplayMessage("Stopping Agent...");
			_Agent.stop();

			// while (_Agent.isRunning())
			// ;

			// FrameGraphics.DisplayMessage("Agent Stopped.");
		}
	}

	public static void interruptAgent() {
		if (_Agent != null) {
			_Agent.interrupt();
		}
	}

	/**
	 * Takes a list of Methods and finds one that accepts the given list of
	 * objects as parameters
	 * 
	 * @param possibleMethods
	 *            The list of Methods to search through
	 * @param parameters
	 *            The array of parameters that the Method should accept
	 * @return The Method found if one was found, otherwise null
	 */
	/*
	 * private static Method FindCompatibleMethods(List<Method>
	 * possibleMethods, Object[] parameters){ Method toRun = null;
	 * 
	 * for(Method m : possibleMethods){ //the number of parameters must be the
	 * same if(parameters.length == m.getParameterTypes().length){ //if no
	 * methods are required, there is nothing else to check if(parameters.length ==
	 * 0) toRun = m; else //check that the list of parameters are of the right
	 * types for(int i = 0; i < parameters.length; i++){
	 * if(m.getParameterTypes()[i].isInstance(parameters[i])) toRun = m; else
	 * toRun = null; }
	 * 
	 * //if a method has been found already, there is no need to keep searching
	 * if(toRun != null) break; } }
	 * 
	 * return toRun; }
	 */

	/**
	 * Converts the given String of values into an array of Objects
	 * 
	 * @param launcher
	 *            The Item used to launch the action, it may be required as a
	 *            parameter
	 * @param values
	 *            A list of space separated String values to convert to objects
	 * @return The created array of Objects
	 */
	public static Object[] CreateObjects(Method method, Frame source,
			Item launcher, String values) {
		// The parameter types that should be created from the given String
		Class[] paramTypes = method.getParameterTypes();

		// if the method has no parameters
		if (paramTypes.length == 0)
			return new Object[0];

		Object[] objects = new Object[paramTypes.length];
		int ind = 0;

		// if the first class in the list is a frame or item, it is the source
		// or launcher
		// length must be at least one if we are still running
		if (paramTypes[0] == Frame.class) {
			objects[0] = source;
			ind = 1;
		} else if (paramTypes[0] == Item.class) {
			objects[0] = launcher;
			ind = 1;
		}

		// if the first class was a frame or item, the second class might be the
		// other
		if (paramTypes.length > 1 && objects[0] != null) {
			if (paramTypes[1] == Frame.class) {
				objects[1] = source;
				ind = 2;
			} else if (paramTypes[1] == Item.class) {
				objects[1] = launcher;
				ind = 2;
			}
		}

		String param;
		// convert the rest of the objects
		for (; ind < objects.length; ind++) {
			// strip off the next value
			param = ParseValue(values);
			values = RemainingParams(values);

			// check if its the last param and combine
			if (values.length() > 0 && ind == objects.length - 1)
				param = param.trim() + ' ' + values.trim();

			// convert the value to an object
			Object o = Conversion.Convert(paramTypes[ind], param);
			objects[ind] = o;
		}

		return objects;

		// strip off the first parameter

		/*
		 * //the list of objects to return List<Object> returnParams = new
		 * ArrayList<Object>(); if(launcher != null)
		 * returnParams.add(launcher);
		 * 
		 * while(param != null){ /* //the parameter is a String
		 * if(param.charAt(0) == '"') returnParams.add(param.substring(1,
		 * param.length() - 1)); else{ //try to parse an int try{ int val =
		 * Integer.parseInt(param); returnParams.add(val);
		 * }catch(NumberFormatException nfe){ System.out.println("Type not found
		 * for: " + param); } }
		 * 
		 * if(param.length() < values.length()) values =
		 * values.substring(param.length() + 1); else values = "";
		 */

		// returnParams.add(KMSConversion.Convert(type, value))
		// get the next value
		/*
		 * param = ParseValue(values); }
		 * 
		 * return returnParams.toArray();
		 */
	}

	/**
	 * Returns a string containing the remaining params after ignoring the first
	 * one.
	 * 
	 * @param params
	 *            a space sparated list of N parameters
	 * @return the remaining N - 1 parameters
	 */
	private static String RemainingParams(String params) {
		if (params.length() == 0)
			return null;

		// remove leading and trailing spaces
		params = params.trim();

		// if there are no more parameters, we are done
		if (params.indexOf(" ") < 0) {
			return "";
		}

		// Check if we have a string parameter
		if (params.charAt(0) == '"') {
			return params.substring(params.indexOf('"', 1) + 1);
		}

		return params.substring(params.indexOf(" "));
	}

	/**
	 * Returns the first value in the space separated String of parameters
	 * passed in. Strings are enclosed in double quotes.
	 * 
	 * @param params
	 *            The String of space separated values
	 * @return The first value in the String
	 */
	private static String ParseValue(String params) {
		if (params.length() == 0)
			return null;

		// remove leading and trailing spaces
		String param = params.trim();

		// if there are no more parameters, we are done
		if (param.indexOf(" ") < 0) {
			return param;
		}

		// Check if we have a string parameter
		if (param.charAt(0) == '"') {
			return param.substring(1, param.indexOf('"', 1));
		}

		return param.substring(0, param.indexOf(" "));
	}

	/**
	 * Separates the name of the given command from any parameters and returns
	 * them
	 * 
	 * @param command
	 *            The String to separate out the Action or Agent name from
	 * @return The name of the Action of Agent with parameters stripped off
	 */
	private static String getName(String command) {
		if (command.indexOf(" ") < 0)
			return command;

		return command.substring(0, command.indexOf(" "));
	}
}
