package org.expeditee.agents;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameUtils;
import org.expeditee.gui.TimeKeeper;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Text;

/**
 * Displays the Tree of Frames starting from the given frame. The delay between
 * frames can be adjusted, either per-frame with the
 * 
 * @DisplayFramePause tag, or mid-stream with the
 * @DisplayTreePause tag, both tags should be followed with the desired delay in
 *                   ms. The default delay between Frames is 50ms.
 * @author jdm18
 * 
 */
public class DisplayTree extends TreeProcessor {

	// delay between frames, in ms
	private long _delay = 50;

	// The shortest delay between frames
	private static final long TIMER_RESOLUTION = 10;

	private long _timeRemaining = 0;

	private int _frameCount = 0;

	private TimeKeeper _timer;

	public DisplayTree(String delay) {
		super();
		try {
			_delay = Long.parseLong(delay);
		} catch (Exception e) {
		}
	}

	public DisplayTree() {
		super();
	}

	@Override
	public boolean initialise(Frame start) {
		_timer = new TimeKeeper();

		// push current frame on to back-stack
		DisplayIO.addToBack(start);

		// MIKE: Doesnt think the line below is needed
		// update the delays (if necessary)
		// delay(start);

		return true;
	}

	@Override
	protected void processFrame(Frame toProcess) {
		FrameUtils.ResponseTimer.restart();
		
		// ignore loops
		if (toProcess != DisplayIO.getCurrentFrame())
			FrameUtils.DisplayFrame(toProcess, false);

		// parse the frame for any pause settings
		delay(toProcess);

		_frameCount++;
	}

	/**
	 * Parses through the given Frame to find any
	 * 
	 * @DisplayTreePause or
	 * @DisplayFramePause tags, and updates the delay or pause as necessary
	 * @param toSearch
	 */
	private void delay(Frame toSearch) {
		// check for change in globaly delay time
		Item delay = ItemUtils.FindTag(toSearch.getItems(),
				"@DisplayTreePause:");
		if (delay != null) {
			try {
				// attempt to read in the delay value
				_delay = Long.parseLong(ItemUtils.StripTag(((Text) delay)
						.getFirstLine(), "@DisplayTreePause:"));
				message("DisplayTree delay changed to: " + _delay + "ms");
			} catch (NumberFormatException nfe) {
				message("Incorrect paramter for DisplayTreePause");
			}

		}

		// check for change in delay for this frame only
		delay = ItemUtils.FindTag(toSearch.getItems(), "@DisplayFramePause:");
		if (delay != null) {
			try {
				// attempt to read in the delay value
				long pause = Long.parseLong(ItemUtils.StripTag(((Text) delay)
						.getFirstLine(), "@DisplayFramePause:"));
				pause(pause);
			} catch (NumberFormatException nfe) {
				message("Incorrect paramter for DisplayFramePause");
			}
		} else
			pause(_delay);
	}

	@Override
	protected void finalise(Frame start) {
		if (_frameCount == 0)
			_frameCount++;
		
		long avg = _timer.getElapsedMillis() / _frameCount;
		String stats = "Frames: " + _frameCount + ", Time: "
				+ _timer.getElapsedStringMillis() + ", Average: " + avg;
		String msg = "Tree successfully displayed. ";

		// return the user to the Frame they started on
		if (!_stop) {
			DisplayIO.Back();
		} else {
			msg = "Agent halted by user. ";
		}

		message(msg + stats);
	}

	/**
	 * Pauses the execution of this Agent for the given time period (in ms.)
	 * This is used for the delay between displaying frames.
	 * 
	 * @param time
	 */
	private void pause(long time) {
		try {
			if (time < 0)
				_timeRemaining = Long.MAX_VALUE;
			else
				_timeRemaining = time;

			while (_timeRemaining > 0) {
				Thread.yield();
				Thread.sleep(TIMER_RESOLUTION);
				_timeRemaining -= TIMER_RESOLUTION;
			}
		} catch (InterruptedException e) {
		}
	}

	@Override
	public void interrupt() {
		_timeRemaining = 0;
	}

	@Override
	public void stop() {
		_timeRemaining = 0;
		super.stop();
	}

}
