package org.expeditee.gui;

import java.awt.AWTException;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Image;
import java.awt.Point;
import java.awt.Robot;
import java.awt.Toolkit;
import java.awt.image.MemoryImageSource;
import java.util.List;
import java.util.Stack;

import javax.swing.JOptionPane;

import org.expeditee.items.Item;
import org.expeditee.items.Picture;
import org.expeditee.stats.SessionStats;

/**
 * This Interface is used by the Frame to control all display input and output.
 * 
 * @author jdm18
 * 
 */
public class DisplayIO {

	private static final int SMALL_CURSOR_SIZE = 16;

	private static final int MEDIUM_CURSOR_SIZE = 32;

	private static final int LARGE_CURSOR_SIZE = 64;

	public static final Color DEFAULT_BACKGROUND = Color.WHITE;

	/**
	 * The color to be used to highlight the linked parent item, when the user
	 * navigates backwards.
	 */
	public static final Color BACK_HIGHLIGHT_COLOR = Color.MAGENTA;

	private static Browser _Browser;

	// The current Frame being displayed on the screen.
	private static Frame _CurrentFrames[] = new Frame[2];

	// Maintains the list of frames visited thus-far for back-tracking
	@SuppressWarnings("unchecked")
	private static Stack<String>[] _VisitedFrames = new Stack[2];

	// used to change the mouse cursor position on the screen
	private static Robot _Robot;

	// public static JMenuBar _MenuBar = new JMenuBar();

	private static boolean _TwinFrames = false;

	/**
	 * The title to display in the Title bar.
	 */
	public static final String TITLE = "Exp08May2008B";

	private DisplayIO() {
	}

	public static void Init(Browser browser) {
		_Browser = browser;
		try {
			_Robot = new Robot();
		} catch (AWTException e) {
			e.printStackTrace();
		}

		Point mouse = _Browser.getMousePosition();
		if (mouse != null) {
			FrameMouseActions.MouseX = mouse.x;
			FrameMouseActions.MouseY = mouse.y;
		}

		_VisitedFrames[0] = new Stack<String>();
		_VisitedFrames[1] = new Stack<String>();
	}

	public static void setTextCursor(int size) {
		if (cursorType == Item.TEXT_CURSOR)
			return;

		cursorType = Item.TEXT_CURSOR;

		// Do some stuff to adjust the cursor size based on the font size
		final int MEDIUM_CURSOR_CUTOFF = 31;
		final int LARGE_CURSOR_CUTOFF = 62;

		int cursorSize = LARGE_CURSOR_SIZE;
		int hotspotPos = 0;
		int start = 0;

		if (size < MEDIUM_CURSOR_CUTOFF) {
			cursorSize = MEDIUM_CURSOR_SIZE;
			start = cursorSize - size - 2;
			hotspotPos = cursorSize - (size + 2) / 4;
		} else if (size < LARGE_CURSOR_CUTOFF) {
			hotspotPos = cursorSize - (size - 5) / 4;
			start = cursorSize - size - 2;
		} else {
			int FIXED_CURSOR_MIN = 77;
			if (size >= FIXED_CURSOR_MIN) {
				hotspotPos = cursorSize - 2;
			} else {
				hotspotPos = size - (FIXED_CURSOR_MIN - cursorSize);
			}
		}

		/*
		 * System.out.println("hs: " + hotspotPos + " start: " + start + " font: " +
		 * size + " cursor: " + (cursorSize - start));
		 */

		int[] pixels = new int[cursorSize * cursorSize];

		for (int i = start; i < cursorSize; i++)
			pixels[i * cursorSize] = pixels[(i * cursorSize) + 1] = 0xFF000000;

		Image image = Toolkit.getDefaultToolkit().createImage(
				new MemoryImageSource(cursorSize, cursorSize, pixels, 0,
						cursorSize));
		Cursor textCursor = Toolkit.getDefaultToolkit().createCustomCursor(
				image, new Point(0, hotspotPos), "textcursor");
		_Browser.setCursor(textCursor);
	}

	/**
	 * Sets the type of cursor the display should be using
	 * 
	 * @param type
	 *            The type of cursor to display, using constants defined in the
	 *            Cursor class.
	 */
	public static void setCursor(int type) {
		// avoid flicker when not changing
		if (type == cursorType || type == Item.UNCHANGED_CURSOR)
			return;

		cursorType = type;

		if (type == Item.HIDDEN_CURSOR) {
			int[] pixels = new int[SMALL_CURSOR_SIZE * SMALL_CURSOR_SIZE];
			Image image = Toolkit.getDefaultToolkit().createImage(
					new MemoryImageSource(SMALL_CURSOR_SIZE, SMALL_CURSOR_SIZE,
							pixels, 0, SMALL_CURSOR_SIZE));
			Cursor transparentCursor = Toolkit.getDefaultToolkit()
					.createCustomCursor(image, new Point(0, 0),
							"invisiblecursor");
			_Browser.setCursor(transparentCursor);
		} else
			_Browser.setCursor(new Cursor(type));
	}

	private static int cursorType = Item.DEFAULT_CURSOR;

	public static int getCursor() {
		return cursorType;
	}

	/**
	 * Moves the mouse cursor to the given x,y coordinates on the screen
	 * 
	 * @param x
	 *            The x coordinate
	 * @param y
	 *            The y coordinate
	 */
	public static void setCursorPosition(int x, int y) {
		setCursorPosition(x, y, true);
	}

	public static void setCursorPosition(int x, int y, boolean forceArrow) {
		int deltax = x - FrameMouseActions.MouseX;
		int deltay = y - FrameMouseActions.MouseY;

		if (_TwinFrames) {
			if (getCurrentSide() == 1)
				x += getMiddle();
		}

		// ensures the cursor doesn't fall behind while waiting for robot
		FrameMouseActions.MouseX = x;
		FrameMouseActions.MouseY = y;

		if (Frame.itemAttachedToCursor()) {
			List<Item> toMove = Frame.FreeItems;
			for (Item move : toMove) {
				move.setPosition(move.getX() + deltax, move.getY() + deltay);
			}
		}

		// cheat
		FrameMouseActions.setForceArrow(forceArrow);
		_Robot.mouseMove((int) _Browser.getContentPane().getLocationOnScreen()
				.getX()
				+ x, (int) _Browser.getContentPane().getLocationOnScreen()
				.getY()
				+ y);
	}

	public static void resetCursorOffset() {
		FrameMouseActions.resetOffset();
	}

	public static void setCursorPosition(Point pos) {
		setCursorPosition(pos.x, pos.y);
	}

	public static void setCursorPosition(Point pos, boolean forceArrow) {
		setCursorPosition(pos.x, pos.y, forceArrow);
	}

	private static boolean _typed = false;

	public static void setKeyTyped(boolean val) {
		_typed = val;
	}

	public static boolean getKeyTyped() {
		return _typed;
	}

	/**
	 * Forces a complete repaint of the GUI and Frame
	 */
	public static void repaint() {
		_Browser.repaint();
	}

	/**
	 * Returns the top item (last added) of the Back-Stack (which is popped off)
	 * 
	 * @return The name of the last Frame added to the back-stack
	 */
	public static String getLastFrame() {
		int side = getCurrentSide();

		if (_VisitedFrames[side].size() > 0)
			return _VisitedFrames[side].pop();
		else
			return null;
	}

	/**
	 * Adds the given Frame to the back-stack
	 * 
	 * @param frame
	 *            The Frame to add
	 */
	public static void addToBack(Frame toAdd) {
		int side = getCurrentSide();

		// do not allow duplicate frames
		if (_VisitedFrames[side].size() > 0)
			if (_VisitedFrames[side].peek().equals(toAdd.getFrameName())) {
				return;
			}

		Item ip = FrameUtils.getCurrentItem();
		if (ip == null)
			_VisitedFrames[side].push(toAdd.getFrameName());
		else
			_VisitedFrames[side].push(toAdd.getFrameName());
	}

	/**
	 * Returns a 'peek' at the end element on the back-stack of the current
	 * side. If the back-stack is empty, null is returned.
	 * 
	 * @return The name of the most recent Frame added to the back-stack, or
	 *         null if the back-stack is empty.
	 */
	public static String peekFromBack() {
		int side = getCurrentSide();

		// check that the stack is not empty
		if (_VisitedFrames[side].size() > 0)
			return _VisitedFrames[side].peek();

		// if the stack is empty, return null
		return null;
	}

	/**
	 * Sets the currently displayed Frame to the given Frame
	 * 
	 * @param frame
	 *            The Frame to display on the screen
	 */
	public static void setCurrentFrame(Frame frame) {
		if (frame == null)
			return;

		if (_TwinFrames) {
			if (_CurrentFrames[0] == null) {
				_CurrentFrames[0] = frame;
				return;
			}
			if (_CurrentFrames[1] == null) {
				_CurrentFrames[1] = frame;
				return;
			}
		}

		// if this is already the current frame
		if (frame == getCurrentFrame()) {
			FrameGraphics.Repaint();
			FrameGraphics.DisplayMessage(frame.getFrameName()
					+ " is already the current frame.");
			return;
		} else {
			SessionStats.AccessedFrame();
		}

		if (_TwinFrames) {
			// if the same frame is being shown in both sides, load a fresh
			// copy
			// from disk
			if (_CurrentFrames[getOppositeSide()] == frame
					|| _CurrentFrames[getOppositeSide()].getOverlays()
							.contains(new Overlay(frame, 1))) {
				FrameIO.SuspendCache();
				frame = FrameIO.LoadFrame(frame.getFrameName());
				FrameIO.ResumeCache();
			}

			// If the frames are the same then the items for the
			// frame that is just about to hide will still be in view
			// so only notify items that they are hidden if the
			// frames differ.
			if (_CurrentFrames[getCurrentSide()] != null
					&& _CurrentFrames[0] != _CurrentFrames[1]) {
				for (Item i : _CurrentFrames[getCurrentSide()].getItems()) {
					i.onParentFameHidden();
				}
			}
			_CurrentFrames[getCurrentSide()] = frame;

			// BROOK
			for (Item i : _CurrentFrames[getCurrentSide()].getItems()) {
				i.onParentFameShown();
			}
		} else {
			// Notifying items on the frame being hidden that they
			// are about to be hidden.
			// ie. Widgets use this method to remove themselves from
			// the JPanel
			if (_CurrentFrames[getCurrentSide()] != null) {
				for (Item i : _CurrentFrames[getCurrentSide()].getItems()) {
					i.onParentFameHidden();
				}
			}

			_CurrentFrames[getCurrentSide()] = frame;
			frame.setMaxSize(FrameGraphics.getMaxFrameSize());

			// Notify items on the frame being displayed that they are in view
			// ie. widgets use this method to add themselves to the content pane
			for (Item i : _CurrentFrames[getCurrentSide()].getItems()) {
				i.onParentFameShown();
			}
		}

		FrameGraphics.Repaint();
	}

	public static void UpdateTitle() {
		StringBuffer title = new StringBuffer(TITLE + " [");

		if (FrameGraphics.isAudienceMode())
			title.append(" - Audience Mode");
		else
			title.append(SessionStats.getShortStats()).append(']');

		_Browser.setTitle(title.toString());
	}

	public static int getCurrentSide() {
		if (_Browser == null)
			return 0;

		if (_TwinFrames
				&& FrameMouseActions.MouseX >= (_Browser.getWidth() / 2)
				&& _CurrentFrames[1] != null)
			return 1;

		if (_CurrentFrames[0] == null && _CurrentFrames[1] != null)
			return 1;

		return 0;
	}

	private static int getOppositeSide() {
		if (getCurrentSide() == 0)
			return 1;

		return 0;
	}

	public static int FrameOnSide(Frame toFind) {
		if (_CurrentFrames[0] == toFind)
			return 0;

		if (_CurrentFrames[1] == toFind)
			return 1;

		return -1;
	}

	/**
	 * Returns the Frame currently being displayed on the screen.
	 * 
	 * @return The Frame currently displayed.
	 */
	public static Frame getCurrentFrame() {
		return _CurrentFrames[getCurrentSide()];
	}

	public static Frame getOppositeFrame() {
		return _CurrentFrames[getOppositeSide()];
	}

	public static Frame[] getFrames() {
		return _CurrentFrames;
	}

	public static int getMiddle() {
		return _Browser.getWidth() / 2;
	}

	public static int getHeight() {
		return _Browser.getHeight();
	}

	/**
	 * Adds the given JMenu to the JMenuBar of the GUI
	 * 
	 * @param menu
	 *            The menu to add
	 */
	/*
	 * public static void addMenu(JMenu menu) { if (menu.getItemCount() == 0)
	 * return;
	 * 
	 * _MenuBar.add(menu); // show progress _Browser.loadedMenu(menu.getText()); }
	 */

	/**
	 * Performs the same action as if the user had clicked on the MenuItem at
	 * the given index in the Menu at the given index.
	 * 
	 * @param menu
	 *            The index of the menu (starts from 0)
	 * @param index
	 *            The index of the Item to simulate a click on (starts from 0)
	 */
	/*
	 * public static void activateMenuItem(int menu, int index) { if
	 * (_MenuBar.getComponentCount() < menu) return;
	 * 
	 * JMenu act = _MenuBar.getMenu(menu);
	 * 
	 * if (act.getComponentCount() < index) return;
	 * 
	 * act.getItem(index).doClick(); }
	 * 
	 * public static int getMenuHeight() { if (_MenuBar.getHeight() < 5) return
	 * 21;
	 * 
	 * return _MenuBar.getHeight() - 5; }
	 */

	/**
	 * Returns the current mouse X coordinate. This coordinate is relative to
	 * the left edge of the frame the mouse is in. It takes into account the
	 * user being in twin frames mode.
	 * 
	 * @return The X coordinate of the mouse.
	 */
	public static int getMouseX() {
		if (_TwinFrames && getMouseY() < FrameGraphics.getMaxSize().height)
			return FrameMouseActions.MouseX % (_Browser.getWidth() / 2);

		return FrameMouseActions.MouseX;
	}

	/**
	 * Returns the current mouse Y coordinate. This is relative to the top of
	 * the frame.
	 * 
	 * @return The Y coordinate of the mouse.
	 */
	public static int getMouseY() {
		return FrameMouseActions.MouseY;
	}

	public static int getRealMouseX() {
		return FrameMouseActions.MouseX;
	}

	public static int getRealMouseY() {
		return FrameMouseActions.MouseY;
	}

	public static void Back() {
		int side = getCurrentSide();

		// there must be a frame to go back to
		if (_VisitedFrames[side].size() < 1) {
			FrameGraphics.DisplayMessage("You are already on the home frame");
			return;
		}

		if (!FrameUtils.LeavingFrame(getCurrentFrame())) {
			FrameGraphics.DisplayMessage("Back operation cancelled");
			return;
		}

		String oldFrame = getCurrentFrame().getFrameName().toLowerCase();

		// do not get a cached version (in case it is in the other window)
		if (isTwinFramesOn())
			FrameIO.SuspendCache();
		Frame frame = FrameIO.LoadFrame(_VisitedFrames[side].pop());
		if (isTwinFramesOn())
			FrameIO.ResumeCache();

		FrameUtils.DisplayFrame(frame, false);
		FrameMouseActions.setHighlightHold(true);

		for (Item i : frame.getItems())
			if (i.getLink() != null
					&& i.getAbsoluteLink().toLowerCase().equals(oldFrame)) {
				i.setSelectedMode(Item.SelectedMode.Normal,
						BACK_HIGHLIGHT_COLOR);
				// check if its an @f item and if so update the buffer
				if (i instanceof Picture) {
					Picture p = (Picture) i;
					p.refresh();
				}
			}

		FrameGraphics.Repaint();
	}

	/**
	 * Toggles the display of frames between TwinFrames mode and Single frame
	 * mode.
	 */
	public static void ToggleTwinFrames() {
		// determine which side is the active side
		int opposite = getOppositeSide();
		int current = getCurrentSide();
		_TwinFrames = !_TwinFrames;

		// if TwinFrames is being turned on
		if (_TwinFrames) {
			// if this is the first time TwinFrames has been toggled on,
			// load
			// the user's first frame
			if (_VisitedFrames[opposite].size() == 0) {
				FrameIO.SuspendCache();
				setCurrentFrame(FrameIO.LoadFrame(UserSettings.FirstFrame));
				FrameIO.ResumeCache();
			} else {
				// otherwise, restore the frame from the side's back-stack
				setCurrentFrame(FrameIO.LoadFrame(_VisitedFrames[opposite]
						.pop()));
			}

			// else, TwinFrames is being turned off
		} else {
			// add the frame to the back-stack
			Frame hiding = _CurrentFrames[opposite];
			FrameUtils.LeavingFrame(hiding);
			_VisitedFrames[opposite].add(hiding.getFrameName());
			_CurrentFrames[opposite] = null;
			_CurrentFrames[current].setMaxSize(FrameGraphics.getMaxFrameSize());
		}
		if (_CurrentFrames[current] != null)
			_CurrentFrames[current].setMaxSize(FrameGraphics.getMaxFrameSize());
		if (_CurrentFrames[opposite] != null)
			_CurrentFrames[opposite]
					.setMaxSize(FrameGraphics.getMaxFrameSize());

		FrameGraphics.Clear();
		FrameGraphics.Repaint();
	}

	public static boolean isTwinFramesOn() {
		return _TwinFrames;
	}

	public static void Reload(int side) {
		if (side < 0)
			return;

		FrameIO.SuspendCache();
		_CurrentFrames[side] = FrameIO.LoadFrame(_CurrentFrames[side]
				.getFrameName());
		FrameIO.ResumeCache();
	}

	public static boolean DisplayConfirmDialog(String message, String title,
			int type, int options, int res) {
		return JOptionPane.showConfirmDialog(_Browser, message, title, options,
				type) == res;
	}

	public static final int RESULT_OK = JOptionPane.OK_OPTION;

	public static final int OPTIONS_OK_CANCEL = JOptionPane.OK_CANCEL_OPTION;

	public static final int TYPE_WARNING = JOptionPane.WARNING_MESSAGE;

	public static void pressMouse(int buttons) {
		_Robot.mousePress(buttons);
	}

	public static void releaseMouse(int buttons) {
		_Robot.mouseRelease(buttons);
	}

	public static void clickMouse(int buttons) {
		_Robot.mousePress(buttons);
		_Robot.mouseRelease(buttons);
	}
}
