package org.expeditee.gui;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.AWTEventListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JComponent;
import javax.swing.JMenuBar;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

import org.expeditee.items.InteractiveWidget;
import org.expeditee.items.Item;

/**
 * The gateway for mouse input; conditionally forwards mouse messages to swing
 * components / expeditee frames for the Browser
 * 
 * @author Brook Novak
 * 
 */
public class MouseEventRouter extends JComponent implements AWTEventListener{

	private static final long serialVersionUID = 1L;

	private JMenuBar _menuBar;

	private Container _contentPane;

	public MouseEventRouter(JMenuBar menuBar, Container contentPane) {

		if (contentPane == null)
			throw new NullPointerException("contentPane");

		// Listen for all AWT events (ensures popups are included)
		Toolkit.getDefaultToolkit().addAWTEventListener(
				this,
				AWTEvent.MOUSE_MOTION_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK
						| AWTEvent.MOUSE_WHEEL_EVENT_MASK);

		this._menuBar = menuBar;
		this._contentPane = contentPane;
	}

	/**
	 * All events for every component in frame are fired to here
	 */
	public void eventDispatched(AWTEvent event) {
		if (event instanceof MouseEvent) {
			routeMouseEvent((MouseEvent) event);
		}
	}

	/**
	 * Forwards the mouse event to its appropriate destination. For events that
	 * belong to Expiditee space the events are consumed and manually routed to
	 * the mouse actions handler.
	 * 
	 * @param e
	 *            The mouse event for any component in the browser frame
	 */
	private void routeMouseEvent(MouseEvent e) {

		// First convert the point to expeditee space
		Point containerPoint = SwingUtilities.convertPoint(e.getComponent(), e
				.getPoint(), _contentPane);

		if (containerPoint.y < 0) { // not in the content pane

			if (_menuBar != null
					&& containerPoint.y + _menuBar.getHeight() >= 0) {
				// The mouse event is over the menu bar.
				// Could handle specially.
			} else {
				// The mouse event is over non-system window
				// decorations, such as the ones provided by
				// the Java look and feel.
				// Could handle specially.
			}

		} else {

			// Check to see if the mouse is over an expeditee item or
			// whether an expeditee item is currently picked up
			boolean forwardToExpiditee = false;

			// Note: all frame.content pane events belong to expeditee
			if (e.getSource() == _contentPane
					|| e.getSource() == Browser._theBrowser
					|| Frame.itemAttachedToCursor()) {
				forwardToExpiditee = true;
			} else if (DisplayIO.getCurrentFrame() != null) { // is mouse over
				// a specific
				// expeditee
				// item?
				// NOTE: Do not use FrameUtils.getCurrentItem() - thats relevent
				// for old mouse position only
				for (Item i : DisplayIO.getCurrentFrame().getItems()) {
					if (i.getPolygon().contains(containerPoint)) {
						forwardToExpiditee = true;
						break;
					}
				}
			}

			if (forwardToExpiditee) {

				// If forwarding to Expiditee ensure that widgets highlighting
				// is enabled
				// do give visual feedback yo users such that Expiditee has
				// focus.
				InteractiveWidget.enableExepiteeHighlighting(true);

				// Ensure that underlying widgets don't get the event
				e.consume();

				// Create artificial mouse event and forward it to expeditee
				MouseEvent expediteeEvent = SwingUtilities.convertMouseEvent(e
						.getComponent(), e, _contentPane);

				switch (e.getID()) {
				case MouseEvent.MOUSE_MOVED:
					FrameMouseActions.getInstance().mouseMoved(expediteeEvent);

					// Ensure that expiditee has focus only if no popups exist
					if (!Browser._theBrowser.getContentPane().isFocusOwner()
							&& !isPopupVisible()) {
						Browser._theBrowser.getContentPane().requestFocus();
					}

					break;
				case MouseEvent.MOUSE_CLICKED:
					FrameMouseActions.getInstance()
							.mouseClicked(expediteeEvent);
					break;
				case MouseEvent.MOUSE_PRESSED:
					FrameMouseActions.getInstance().ProccessMousePressedEvent(
							expediteeEvent, e.getModifiersEx());
					break;
				case MouseEvent.MOUSE_RELEASED:
					FrameMouseActions.getInstance().mouseReleased(
							expediteeEvent);
					break;
				case MouseEvent.MOUSE_WHEEL:
					FrameMouseActions.getInstance().mouseWheelMoved(
							(MouseWheelEvent) expediteeEvent);
					break;
				case MouseEvent.MOUSE_ENTERED:
					FrameMouseActions.getInstance()
							.mouseEntered(expediteeEvent);
					break;
				case MouseEvent.MOUSE_EXITED:
					FrameMouseActions.getInstance().mouseExited(expediteeEvent);
					break;
				case MouseEvent.MOUSE_DRAGGED:
					FrameMouseActions.getInstance()
							.mouseDragged(expediteeEvent);
					break;
				}

			} else {

				// If forwarding to swing ensure that widgets are not
				// highlighted
				// to give visual feedback yo users such that swing has focus.
				InteractiveWidget.enableExepiteeHighlighting(false);

				// Also bring expideditee behaviour to swing: Auto-focus on
				// component
				// whenever the mouse moves over it.
				if (e.getID() == MouseEvent.MOUSE_MOVED) {
					if (e.getSource() instanceof Component) {
						Component target = (Component) e.getSource();
						if (!target.isFocusOwner()) {
							target.requestFocus();
						}
					}
				}
			}
		}
	}

	public static boolean isPopupVisible() {
		return isPopupVisible(Browser._theBrowser.getLayeredPane());
	}

	private static boolean isPopupVisible(Container parent) {

		for (Component c : parent.getComponents()) {

			if (c instanceof JPopupMenu && ((JPopupMenu) c).isVisible()) {
				return true;

			} else if (c instanceof Container
					&& c != Browser._theBrowser.getContentPane()) {
				if (isPopupVisible((Container) c))
					return true;
			}
		}

		return false;
	}
}